unit UnitFuncCurve;

interface

uses
   Windows, UnitFunc ;

var
   tLns  : array of TPointD ;    // ϊAf[^
   tLnsN : integer ;

   procedure TransLnsClear ;
   procedure TransLinesToLines(num:integer;vx,vy:array of double;t:TMatrix);
   procedure TransCircleToLines(sp:integer;px,py,pr:double;t:TMatrix);
   procedure TransArcToLines(sp,dir:integer;px,py,pr,sa,ea:double;t:TMatrix);
   procedure TransEllipseToLines(sp:integer;px,py,rx,ry,ra:double;t:TMatrix);
   procedure TransEllipseArcToLines(sp,dir:integer;px,py,rx,ry,ra,sa,ea:double;t:TMatrix);
   procedure TransBezierToLines(sp,num:integer;vx,vy:array of double;t:TMatrix);
   procedure TransBsplineToLines(oc,sp,num:integer;vx,vy:array of double;t:TMatrix);
   procedure TransSplineToLines(stp,oc,sp,num:integer;av:double;vx,vy:array of double;t:TMatrix);
   procedure TransClothoidToLines(dir,sp:integer;bx,by,av,an,l1,l2:double;t:TMatrix);
   { --- sub ---
   procedure TransBezierSub(var xp,yp:double;t,x1,x2,x3,x4, y1,y2,y3,y4:double);
   procedure TransBsplineSub(var xp,yp:double;t,x1,x2,x3,x4, y1,y2,y3,y4:double);
   procedure TransSplineSub1(var xp,yp:double;t,av,x0,x1,x2,x3, y0,y1,y2,y3:double);
   procedure TransSplineSub2(tn,nmax:integer;t:double;var tv,bn:array of double);
   procedure TransSplineSub3(tn,nmax,kai:integer;t:double;var tv,bn1,bn2:array of double);
   procedure TransClothoidSub1(var x,y:double;p:double);
   procedure TransClothoidSub2(var x,y:double;A,l:double);
   }
   procedure TransTenMarkerToLines(mc:integer;px,py,an,sc:double;t:TMatrix);
   procedure TransArrowToLines(ac:integer;px,py,an,sc:double;t:TMatrix);

implementation

// ϊAf[^̃NA
procedure TransLnsClear ;
begin
   tLns  := nil ;
   tLnsN := 0 ;
end;

// ܐAɕϊ
procedure TransLinesToLines(num:integer;vx,vy:array of double;t:TMatrix);
var
   i,l : integer ;
   px,py : double ;
begin
   try
      tLnsN := num + 1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   l := 0;
   for i:=0 to num-1 do begin
      MtxXY(px,py, vx[i],vy[i], t);
      if (l > 0) then begin
         if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
            tLns[l].x := px ;
            tLns[l].y := py ;
            Inc(l);
         end;
      end
      else begin
         tLns[l].x := px ;
         tLns[l].y := py ;
         Inc(l);
      end;
   end;
   tLnsN := l ;
   SetLength(tLns, tLnsN);
end;

// ~Aɕϊ
procedure TransCircleToLines(sp:integer;px,py,pr:double;t:TMatrix);
var
   j : integer ;
   a,aa, x1,y1 : double ;
begin
   try
      tLnsN := sp + 1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   a  := 2.0*Pi/sp ;
   aa := 0.0 ;
   for j:=0 to sp do begin
      x1 := px + pr*Cos(aa);
      y1 := py + pr*Sin(aa);
      MtxXY(x1,y1, x1,y1, t);
      tLns[j].x := x1 ;
      tLns[j].y := y1 ;
      aa := aa + a;
   end;
end;

// ~ʂAɕϊ
procedure TransArcToLines(sp,dir:integer;px,py,pr,sa,ea:double;t:TMatrix);
var
   j : integer ;
   a,aa, x1,y1 : double ;
begin
   try
      tLnsN := sp + 1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   if (dir <> 0) then swapD(sa,ea);
   aa := ea - sa ;
   if (aa < 0.0) then aa := aa + 360.0 ;
   if (dir <> 0) then begin
      swapD(sa,ea);
      aa := -aa ;
   end;

   a  := aa/180.0*Pi/sp;
   aa := sa/180.0*Pi ;
   for j:=0 to sp do begin
      x1 := px + pr*Cos(aa);
      y1 := py + pr*Sin(aa);
      MtxXY(x1,y1, x1,y1, t);
      tLns[j].x := x1 ;
      tLns[j].y := y1 ;
      aa := aa + a;
   end;
end;

// ȉ~Aɕϊ
procedure TransEllipseToLines(sp:integer;px,py,rx,ry,ra:double;t:TMatrix);
var
   j : integer ;
   a,aa, x1,y1 : double ;
   tr,tr1 : TMatrix ;
begin
   try
      tLnsN := sp + 1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   MtxMove(tr, -px,-py);
   MtxRotD(tr1,ra);
   tr := MtxMult(tr,tr1) ;
   MtxMove(tr1, px,py);
   tr := MtxMult(tr,tr1) ;
   t  := MtxMult(tr,t) ;

   a  := 2.0*Pi/sp ;
   aa := 0.0 ;
   for j:=0 to sp do begin
      x1 := px + rx*Cos(aa);
      y1 := py + ry*Sin(aa);
      MtxXY(x1,y1, x1,y1, t);
      tLns[j].x := x1 ;
      tLns[j].y := y1 ;
      aa := aa + a;
   end;
end;

// ȉ~ʂAɕϊ
procedure TransEllipseArcToLines(sp,dir:integer;px,py,rx,ry,ra,sa,ea:double;t:TMatrix);
var
   j : integer ;
   a,aa, x1,y1 : double ;
   tr,tr1 : TMatrix ;
begin
   try
      tLnsN := sp + 1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   MtxMove(tr, -px,-py);
   MtxRotD(tr1,ra);
   tr := MtxMult(tr,tr1) ;
   MtxMove(tr1, px,py);
   tr := MtxMult(tr,tr1) ;
   t  := MtxMult(tr,t) ;

   if (dir <> 0) then swapD(sa,ea);
   aa := ea - sa ;
   if (aa < 0.0) then aa := aa + 360.0 ;
   if (dir <> 0) then begin
      swapD(sa,ea);
      aa := -aa ;
   end;

   a  := aa/180.0*Pi/sp ;
   aa := sa/180.0*Pi ;
   for j:=0 to sp do begin
      x1 := px + rx*Cos(aa);
      y1 := py + ry*Sin(aa);
      MtxXY(x1,y1, x1,y1, t);
      tLns[j].x := x1 ;
      tLns[j].y := y1 ;
      aa := aa + a;
   end;
end;

// xWFȐ̕ϊ@Tu
procedure TransBezierSub(var xp,yp:double;t,x1,x2,x3,x4, y1,y2,y3,y4:double);
var
   t0,t1,t2,t3 : double ;
begin
   t0 := (1.0-t)*(1.0-t)*(1.0-t);
   t1 := 3.0*t*(1.0-t)*(1.0-t);
   t2 := 3.0*(1.0-t)*t*t;
   t3 := t*t*t;
   xp := t0*x1 + t1*x2 + t2*x3 + t3*x4 ;
   yp := t0*y1 + t1*y2 + t2*y3 + t3*y4 ;
end;

// xWFȐAɕϊ
procedure TransBezierToLines(sp,num:integer;vx,vy:array of double;t:TMatrix);
var
   i,j,k : integer ;
   tt,x1,y1 : double ;
begin
   try
      tLnsN := ((num-1) div 3)*(sp+1)+1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   i := 0 ;
   for j:=0 to ((num-1) div 3)-1 do begin
      tt := 0.0;
      TransBezierSub(x1,y1, tt,vx[j*3],vx[j*3+1],vx[j*3+2],vx[j*3+3],
                               vy[j*3],vy[j*3+1],vy[j*3+2],vy[j*3+3]);
      MtxXY(x1,y1, x1,y1, t);
      if (i > 0) then begin
         if not(SamePosCheck(x1,y1, tLns[i-1].x,tLns[i-1].y)) then begin
            tLns[i].x := x1 ;
            tLns[i].y := y1 ;
            Inc(i);
         end;
      end
      else begin
         tLns[i].x := x1 ;
         tLns[i].y := y1 ;
         Inc(i);
      end;
      for k:=1 to sp do begin
         tt := tt + 1.0/sp ;
         TransBezierSub(x1,y1, tt,vx[j*3],vx[j*3+1],vx[j*3+2],vx[j*3+3],
                                  vy[j*3],vy[j*3+1],vy[j*3+2],vy[j*3+3]);
         MtxXY(x1,y1, x1,y1, t);
         if (i > 0) then begin
            if not(SamePosCheck(x1,y1, tLns[i-1].x,tLns[i-1].y)) then begin
               tLns[i].x := x1 ;
               tLns[i].y := y1 ;
               Inc(i);
            end;
         end
         else begin
            tLns[i].x := x1 ;
            tLns[i].y := y1 ;
            Inc(i);
         end;
      end;
   end;
   tLnsN := i ;
   SetLength(tLns, tLnsN);
end;

// aXvCȐ̕\@Tu
procedure TransBsplineSub(var xp,yp:double;t,x1,x2,x3,x4, y1,y2,y3,y4:double);
var
   t0,t1,t2,t3 : double ;
begin
   t0 := (1.0-t)*(1.0-t)*(1.0-t)/6.0;
   t1 := 0.5*t*t*t - t*t + 2.0/3.0;
   t2 :=-0.5*t*t*t + 0.5*t*t + 0.5*t + 1.0/6.0;
   t3 := t*t*t/6.0;
   xp := t0*x1 + t1*x2 + t2*x3 + t3*x4 ;
   yp := t0*y1 + t1*y2 + t2*y3 + t3*y4 ;
end;

// aXvCȐAɕϊ
procedure TransBsplineToLines(oc,sp,num:integer;vx,vy:array of double;t:TMatrix);
var
   i,j,l : integer ;
   tt,x1,y1,x2,y2,x3,y3,x4,y4,px,py : double ;
begin
   try
      tLnsN := num*(sp+1)+1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   l := 0 ;
   i := 0;
   while(i < num) do begin
      if (i = 0) then begin
         if (oc = 1) then begin
            // JȐ
            x1 := vx[0];
            y1 := vy[0];
         end
         else begin
            // Ȑ
            x1 := vx[num-2];
            y1 := vy[num-2];
         end;
      end
      else begin
         x1 := vx[i-1];
         y1 := vy[i-1];
      end;
      x2 := vx[i];
      y2 := vy[i];
      if (i >= num-1) then begin
         if (oc = 1)or(num < 2) then begin
            // JȐ
            x3 := vx[num-1];
            y3 := vy[num-1];
         end
         else begin
            // Ȑ
            x3 := vx[1];
            y3 := vy[1];
         end;
      end
      else begin
         x3 := vx[i+1];
         y3 := vy[i+1];
      end;
      if (i >= num-2) then begin
         if (oc = 1)or(num < 3) then begin
            // JȐ
            x4 := vx[num-1];
            y4 := vy[num-1];
         end
         else begin
            // Ȑ
            x4 := vx[i-num+3];
            y4 := vy[i-num+3];
         end;
      end
      else begin
         x4 := vx[i+2];
         y4 := vy[i+2];
      end;
      //
      tt  := 0.0;
      TransBsplineSub(px,py, tt,x1,x2,x3,x4, y1,y2,y3,y4);
      if (i = 0)and(oc = 1) then begin
         px := vx[0];
         py := vy[0];
         MtxXY(px,py, px,py, t);
         if (l > 0) then begin
            if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
               tLns[l].x := px ;
               tLns[l].y := py ;
               Inc(l);
            end;
         end
         else begin
            tLns[l].x := px ;
            tLns[l].y := py ;
            Inc(l);
         end;
      end;
      for j:=1 to sp do begin
         tt := tt + 1.0/sp;
         TransBsplineSub(px,py, tt,x1,x2,x3,x4, y1,y2,y3,y4);
         MtxXY(px,py, px,py, t);
         if (l > 0) then begin
            if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
               tLns[l].x := px ;
               tLns[l].y := py ;
               Inc(l);
            end;
         end
         else begin
            tLns[l].x := px ;
            tLns[l].y := py ;
            Inc(l);
         end;
      end;
      i := i + 1;
   end;
   tLnsN := l ;
   SetLength(tLns, tLnsN);
end;

// XvCȐ̕\@TuP
//    J[fBiXvC
procedure TransSplineSub1(var xp,yp:double;t,av,x0,x1,x2,x3, y0,y1,y2,y3:double);
var
   t0,t1,t2,t3,dx1,dy1,dx2,dy2 : double ;
begin
   t0 := 2.0*t*t*t - 3.0*t*t + 1.0;
   t1 :=-2.0*t*t*t + 3.0*t*t;
   t2 := t*t*t - 2.0*t*t + t;
   t3 := t*t*t - t*t;
   dx1 := av*(x2-x0);
   dy1 := av*(y2-y0);
   dx2 := av*(x3-x1);
   dy2 := av*(y3-y1);
   xp := t0*x1 + t1*x2 + t2*dx1 + t3*dx2 ;
   yp := t0*y1 + t1*y2 + t2*dy1 + t3*dy2 ;
end;

// XvCȐ̕\@TuQ
//    de boor-cox̑QɂBXvC֐̌vZ
procedure TransSplineSub2(tn,nmax:integer;t:double;var tv,bn:array of double);
var
   i,j,k : integer ;
   a0,a1,a2 : double ;
begin
   k := 0;
   for i:=0 to nmax-1 do
      bn[i] := 0.0 ;
   for i:=0 to nmax-1 do begin
      if (tv[i]-LIMIT8 <= t)and(t < tv[i+1]-LIMIT8) then begin
         bn[i] := 1.0 ;
         k := i ;
      end;
   end;
   if (tv[nmax-1]-LIMIT8 <= t)and(t <= tv[nmax]+LIMIT8) then begin
      bn[nmax-1] := 1.0 ;
      k := nmax-1 ;
   end;

   for i:=1 to 3 do begin
      for j:=0 to nmax-1 do
         bn[j + nmax*i] := 0.0 ;
      for j:=k-i to k do begin
         a1 := 0.0;
         a2 := 0.0;
         a0 := tv[j+(i+1)-1] - tv[j] ;
         if (Abs(a0) > LIMIT8) then
            a1 := (t - tv[j]) * bn[j + nmax*(i-1)] / a0 ;
         a0 := tv[j+(i+1)] - tv[j+1] ;
         if (Abs(a0) > LIMIT8) then
            a2 := (tv[j+(i+1)] - t) * bn[j+1 + nmax*(i-1)] / a0 ;
         bn[j + nmax*i] := a1 + a2 ;
      end;
   end;
end;

// XvCȐ̕\@TuR
//    de boor-cox̑QɂBXvC֐̂W̌vZ
procedure TransSplineSub3(tn,nmax,kai:integer;t:double;var tv,bn1,bn2:array of double);
var
   i,j : integer ;
   a0,a1,a2 : double ;
begin
   for j:=0 to nmax-1 do
      bn2[j] := 0.0 ;

   for i:=1 to (kai-1) do begin
      for j:=0 to nmax-1 do begin
         a1 := 0.0;
         a2 := 0.0;
         a0 := tv[j+(i+1)-1] - tv[j] ;
         if (Abs(a0) > LIMIT8) then
            a1 := bn1[j + nmax*(i-1)] / a0 ;
         a0 := tv[j+(i+1)] - tv[j+1] ;
         if (Abs(a0) > LIMIT8) then
            a2 := bn1[j+1 + nmax*(i-1)] / a0 ;
         bn2[j + nmax*i] := (i+1.0 - 1.0)*( a1 - a2 );
      end;
   end;
end;

// XvCȐAɕϊ
procedure TransSplineToLines(stp,oc,sp,num:integer;av:double;vx,vy:array of double;t:TMatrix);
var
   i,j,l : integer ;
   tt,x0,y0,x1,y1,x2,y2,x3,y3,px,py : double ;
   //
   m1,m2 : integer ;
   qx,qy : array of double ;
   //
   qn  : integer ;      // _̐
   jisu: integer ;      // 
   kai : integer ;      // K
   tn  : integer ;         // mbgxNg̐
   tv : array of double ;  // mbgxNg
   nmax : integer ;        // ֐̐
   bn : array of double ;  // ֐
   bn1: array of double ;  // ֐̂PW
   bn2: array of double ;  // ֐̂QW
   d : double ;
   maxq,jj : integer ;
   q,x,y : array of double ;
begin
   try
      tLnsN := num*(sp+1)+1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   l := 0 ;
   Case(stp)of
   1: begin
         // J[fBiXvCȐ
         for i:=0 to num-2 do begin
            if (i > 0) then begin
               x0 := vx[i-1];
               y0 := vy[i-1];
            end
            else begin
               if (oc = 1) then begin
                  // JȐ
                  x0 := vx[0];
                  y0 := vy[0];
               end
               else begin
                  // Ȑ
                  x0 := vx[num-2];
                  y0 := vy[num-2];
               end;
            end;
            x1 := vx[i];
            y1 := vy[i];
            x2 := vx[i+1];
            y2 := vy[i+1];
            if (i < (num-2)) then begin
               x3 := vx[i+2];
               y3 := vy[i+2];
            end
            else begin
               if (oc = 1) then begin
                  // JȐ
                  x3 := vx[i+1];
                  y3 := vy[i+1];
               end
               else begin
                  // Ȑ
                  x3 := vx[1];
                  y3 := vy[1];
               end;
            end;
            //
            tt := 0.0;
            TransSplineSub1(px,py, tt,av, x0,x1,x2,x3, y0,y1,y2,y3);
            MtxXY(px,py, px,py, t);
            if (l > 0) then begin
               if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
                  tLns[l].x := px ;
                  tLns[l].y := py ;
                  Inc(l);
               end;
            end
            else begin
               tLns[l].x := px ;
               tLns[l].y := py ;
               Inc(l);
            end;
            for j:=1 to sp do begin
               tt := tt + 1.0/sp;
               TransSplineSub1(px,py, tt,av, x0,x1,x2,x3, y0,y1,y2,y3);
               MtxXY(px,py, px,py, t);
               if (l > 0) then begin
                  if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
                     tLns[l].x := px ;
                     tLns[l].y := py ;
                     Inc(l);
                  end;
               end
               else begin
                  tLns[l].x := px ;
                  tLns[l].y := py ;
                  Inc(l);
               end;
            end;
         end;
      end;
   2: begin
         // xWF
         // r̐_Zo邽߂̍s𐶐
         m1 := (num-1)*2+1;
         m2 := (num-1)*2;
         SetLength(qx,m1*m2);
         SetLength(qy,m1*m2);
         for i:=0 to m1*m2-1 do
            qx[i] := 0.0;
         for i:=0 to num-3 do begin
            qx[(i*2  )*m1+i*2+1] := 1.0;
            qx[(i*2  )*m1+i*2+2] := 1.0;
            qx[(i*2+1)*m1+i*2  ] := 1.0;
            qx[(i*2+1)*m1+i*2+1] :=-2.0;
            qx[(i*2+1)*m1+i*2+2] := 2.0;
            qx[(i*2+1)*m1+i*2+3] :=-1.0;
         end;
         if (oc = 1) then begin
            // JȐ
            qx[((num-2)*2  )*m1  ]           := 2.0;
            qx[((num-2)*2  )*m1+1]           :=-1.0;
            qx[((num-2)*2+1)*m1+(num-2)*2  ] :=-1.0;
            qx[((num-2)*2+1)*m1+(num-2)*2+1] := 2.0;
         end
         else begin
            // Ȑ
            qx[((num-2)*2  )*m1  ]           := 1.0;
            qx[((num-2)*2  )*m1+(num-2)*2+1] := 1.0;
            qx[((num-2)*2+1)*m1  ]           := 2.0;
            qx[((num-2)*2+1)*m1+1]           :=-1.0;
            qx[((num-2)*2+1)*m1+(num-2)*2  ] := 1.0;
            qx[((num-2)*2+1)*m1+(num-2)*2+1] :=-2.0;
         end;
         for i:=0 to m1*m2-1 do
            qy[i] := qx[i] ;
         // sɍWlZbg
         for i:=0 to num-3 do begin
            qx[(i*2+1)*m1-1] := 2.0*vx[i+1];
            qy[(i*2+1)*m1-1] := 2.0*vy[i+1];
         end;
         if (oc = 1) then begin
            // JȐ
            qx[((num-2)*2+1)*m1-1] := vx[0];
            qy[((num-2)*2+1)*m1-1] := vy[0];
            qx[((num-2)*2+2)*m1-1] := vx[num-1];
            qy[((num-2)*2+2)*m1-1] := vy[num-1];
         end
         else begin
            // Ȑ
            qx[((num-2)*2+1)*m1-1] := 2.0*vx[num-1];
            qy[((num-2)*2+1)*m1-1] := 2.0*vy[num-1];
         end;
         // s
         gauss(m1,m2,qx);
         gauss(m1,m2,qy);
         // }
         for i:=0 to num-2 do begin
            x0 := vx[i];
            y0 := vy[i];
            x1 := qx[(i*2+1)*m1-1];
            y1 := qy[(i*2+1)*m1-1];
            x2 := qx[(i*2+2)*m1-1];
            y2 := qy[(i*2+2)*m1-1];
            x3 := vx[i+1];
            y3 := vy[i+1];
            tt := 0.0;
            TransBezierSub(px,py, tt,x0,x1,x2,x3, y0,y1,y2,y3);
            MtxXY(px,py, px,py, t);
            if (l > 0) then begin
               if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
                  tLns[l].x := px ;
                  tLns[l].y := py ;
                  Inc(l);
               end;
            end
            else begin
               tLns[l].x := px ;
               tLns[l].y := py ;
               Inc(l);
            end;
            for j:=1 to sp do begin
               tt := tt + 1.0/sp;
               TransBezierSub(px,py, tt,x0,x1,x2,x3, y0,y1,y2,y3);
               MtxXY(px,py, px,py, t);
               if (l > 0) then begin
                  if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
                     tLns[l].x := px ;
                     tLns[l].y := py ;
                     Inc(l);
                  end;
               end
               else begin
                  tLns[l].x := px ;
                  tLns[l].y := py ;
                  Inc(l);
               end;
            end;
         end;
         qx := nil;
         qy := nil;
      end;
   3: begin
         // NUBSimjtH[aXvCȐj
         qn   := num + 2 ;
         jisu := 3;
         kai  := jisu + 1;
         tn   := qn + kai ;
         // mbgxNg
         SetLength(tv,tn);
         tv[0] := 0.0 ;
         tv[1] := 0.0 ;
         tv[2] := 0.0 ;
         tv[3] := 0.0 ;
         tt := 0.0 ;
         for i:=0 to num-2 do begin
            d := Dist(vx[i],vy[i], vx[i+1],vy[i+1]);
            tt := tt + d ;
            tv[4+i] := tt ;
         end;
         tv[3+num] := tv[2+num] ;
         tv[4+num] := tv[2+num] ;
         tv[5+num] := tv[2+num] ;
         //
         nmax := qn ;
         SetLength(bn , nmax*4);
         SetLength(bn1, nmax*4);
         SetLength(bn2, nmax*4);
         SetLength(qx, (qn)*(qn+1));
         SetLength(qy, (qn)*(qn+1));
         for i:=0 to ((qn)*(qn+1)-1) do begin
            qx[i] := 0.0;
            qy[i] := 0.0;
         end;
         // aXvC֐@@ʉߓ__
         for i:=0 to num-1 do begin
            tt := tv[i+jisu] ;
            TransSplineSub2(tn,nmax,tt, tv,bn);
            for j:=0 to nmax-1 do begin
               qx[j +(i)*(qn+1)] := bn[j + nmax*3] ;
               qy[j +(i)*(qn+1)] := bn[j + nmax*3] ;
            end;
            qx[nmax +(i)*(qn+1)] := vx[i];
            qy[nmax +(i)*(qn+1)] := vy[i];
         end;
         if (oc = 1) then begin
            // JȐ[̋ȗ=0Ql=0
            i := 0;
            tt:= tv[i+jisu] ;
            TransSplineSub2(tn,nmax,tt, tv,bn);
            TransSplineSub3(tn,nmax,kai,tt, tv,bn ,bn1);
            TransSplineSub3(tn,nmax,kai,tt, tv,bn1,bn2);
            for j:=0 to nmax-1 do begin
               qx[j +(num)*(qn+1)] := bn2[j + nmax*3] ;
               qy[j +(num)*(qn+1)] := bn2[j + nmax*3] ;
            end;
            qx[nmax +(num)*(qn+1)] := 0.0 ;
            qy[nmax +(num)*(qn+1)] := 0.0 ;
            i := num-1;
            tt:= tv[i+jisu] ;
            TransSplineSub2(tn,nmax,tt, tv,bn);
            TransSplineSub3(tn,nmax,kai,tt, tv,bn ,bn1);
            TransSplineSub3(tn,nmax,kai,tt, tv,bn1,bn2);
            for j:=0 to nmax-1 do begin
               qx[j +(num+1)*(qn+1)] := bn2[j + nmax*3] ;
               qy[j +(num+1)*(qn+1)] := bn2[j + nmax*3] ;
            end;
            qx[nmax +(num+1)*(qn+1)] := 0.0 ;
            qy[nmax +(num+1)*(qn+1)] := 0.0 ;
         end
         else begin
            // Ȑ[̋ȗ=0Ql̍=0
            //       @[̐ڐ=0Pl̍=0
            i := 0;
            tt:= tv[i+jisu] ;
            TransSplineSub2(tn,nmax,tt, tv,bn);
            TransSplineSub3(tn,nmax,kai,tt, tv,bn ,bn1);
            TransSplineSub3(tn,nmax,kai,tt, tv,bn1,bn2);
            for j:=0 to nmax-1 do begin
               qx[j +(num)  *(qn+1)] := bn2[j + nmax*3] ;
               qy[j +(num)  *(qn+1)] := bn2[j + nmax*3] ;
               qx[j +(num+1)*(qn+1)] := bn1[j + nmax*3] ;
               qy[j +(num+1)*(qn+1)] := bn1[j + nmax*3] ;
            end;
            i := num-1;
            tt:= tv[i+jisu] ;
            TransSplineSub2(tn,nmax,tt, tv,bn);
            TransSplineSub3(tn,nmax,kai,tt, tv,bn ,bn1);
            TransSplineSub3(tn,nmax,kai,tt, tv,bn1,bn2);
            for j:=0 to nmax-1 do begin
               qx[j +(num)  *(qn+1)] := qx[j +(num)  *(qn+1)] - bn2[j + nmax*3] ;
               qy[j +(num)  *(qn+1)] := qy[j +(num)  *(qn+1)] - bn2[j + nmax*3] ;
               qx[j +(num+1)*(qn+1)] := qx[j +(num+1)*(qn+1)] - bn1[j + nmax*3] ;
               qy[j +(num+1)*(qn+1)] := qy[j +(num+1)*(qn+1)] - bn1[j + nmax*3] ;
            end;
            qx[nmax +(num)  *(qn+1)] := 0.0 ;
            qy[nmax +(num)  *(qn+1)] := 0.0 ;
            qx[nmax +(num+1)*(qn+1)] := 0.0 ;
            qy[nmax +(num+1)*(qn+1)] := 0.0 ;
         end;
         // s񎮂
         gauss(qn+1,qn,qx);
         gauss(qn+1,qn,qy);
         // eԖɕ`s
         maxq := (jisu+1) + kai ; // 1ԕ
         SetLength(q, maxq);
         SetLength(x, jisu+1);
         SetLength(y, jisu+1);
         for j:=0 to (num-2) do begin
            // mbgxNg
            for i:=0 to (maxq-1) do
               q[i] := tv[i+j];
            // _
            for i:=0 to jisu do begin
               x[i] := qx[qn+(i+j)*(qn+1)] ;
               y[i] := qy[qn+(i+j)*(qn+1)] ;
            end;
            tt  := q[kai-1] ;
            for i:=0 to sp-1 do begin
               px := 0.0 ;
               py := 0.0 ;
               for jj:=0 to kai-1 do begin
                  TransSplineSub2(maxq,kai+2,tt, q,bn);
                  px := px + bn[jj + (kai+2)*3]*x[jj] ;
                  py := py + bn[jj + (kai+2)*3]*y[jj] ;
               end;
               MtxXY(px,py, px,py, t);
               if (l > 0) then begin
                  if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
                     tLns[l].x := px ;
                     tLns[l].y := py ;
                     Inc(l);
                  end;
               end
               else begin
                  tLns[l].x := px ;
                  tLns[l].y := py ;
                  Inc(l);
               end;
               tt := tt + (q[kai]-q[kai-1])*1.0/sp ;
            end;
            px := vx[j+1];
            py := vy[j+1];
            MtxXY(px,py, px,py, t);
            if (l > 0) then begin
               if not(SamePosCheck(px,py, tLns[l-1].x,tLns[l-1].y)) then begin
                  tLns[l].x := px ;
                  tLns[l].y := py ;
                  Inc(l);
               end;
            end
            else begin
               tLns[l].x := px ;
               tLns[l].y := py ;
               Inc(l);
            end;
         end;
         //
         q := nil;
         x := nil;
         y := nil;
         tv := nil;
         bn := nil;
         bn1:= nil;
         bn2:= nil;
         qx := nil;
         qy := nil;
      end;
   end;
   tLnsN := l ;
   SetLength(tLns, tLnsN);
end;

// N\ChȐ̕\ TuP
//    N\ChȐ̍W𓾂錳
procedure TransClothoidSub1(var x,y:double;p:double);
begin
   x := Cos(Pi*p*p/2.0);
   y := Sin(Pi*p*p/2.0);
end;

// N\ChȐ̕\ TuQ
//    N\ChȐ̍W𓾂@Vv\𗘗p
procedure TransClothoidSub2(var x,y:double;A,l:double);
const
   CALC_SEP = 100 ;     // vZ̕
var
   aa,u,h,sx1,sy1,sx2,sy2,p,w1,w2,w3,w4 : double ;
   i : integer ;
begin
   aa := A*Sqrt(Pi);
   u  := l/aa ;
   h  := u/CALC_SEP;
   sx1:= 0.0;
   sy1:= 0.0;
   sx2:= 0.0;
   sy2:= 0.0;
   i  := 1;
   while(i <= (CALC_SEP-1)) do begin
      p := i*h ;
      TransClothoidSub1(w1,w2,p);
      sx1 := sx1 + 4.0*w1 ;
      sy1 := sy1 + 4.0*w2 ;
      i := i + 2;
   end;
   i  := 2;
   while(i <= (CALC_SEP-2)) do begin
      p := i*h ;
      TransClothoidSub1(w1,w2,p);
      sx2 := sx2 + 2.0*w1 ;
      sy2 := sy2 + 2.0*w2 ;
      i := i + 2;
   end;
   TransClothoidSub1(w1,w2,0.0);
   TransClothoidSub1(w3,w4,u);
   x := aa*h*(w1 + sx1 + sx2 + w3)/3.0 ;
   y := aa*h*(w2 + sy1 + sy2 + w4)/3.0 ;
end;

// N\ChȐAɕϊ
procedure TransClothoidToLines(dir,sp:integer;bx,by,av,an,l1,l2:double;t:TMatrix);
var
   tr,tr1 : TMatrix ;
   i,k : integer ;
   ls,lt,ll,x1,y1,x2,y2,mx,my,aa : double ;
begin
   try
      tLnsN := sp + 1 ;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   MtxRotD(tr, an);
   MtxMove(tr1, bx,by);
   tr := MtxMult(tr,tr1) ;
   t  := MtxMult(tr,t) ;

   k := 0;
   ls:= (l2-l1)/sp ;
   lt:= l1 ;
   ll:= 0.0;
   TransClothoidSub2(x1,y1,av,lt);
   if (dir = 1) then y1 := -y1;
   MtxXY(mx,my, x1,y1, t);
   tLns[k].x := mx ;
   tLns[k].y := my ;
   Inc(k);
   for i:=1 to sp do begin
      lt := lt + ls;
      TransClothoidSub2(x2,y2,av,lt);
      if (dir = 1) then y2 := -y2;
      if (i = sp) then begin
         // ŏI̒̕␳
         ll := Abs(l2-l1) - ll ;
         aa := Angle(x2-x1,y2-y1);
         x2 := x1 + ll*Cos(aa);
         y2 := y1 + ll*Sin(aa);
      end;
      MtxXY(mx,my, x2,y2, t);
      if not(SamePosCheck(mx,my, tLns[k-1].x,tLns[k-1].y)) then begin
         tLns[k].x := mx ;
         tLns[k].y := my ;
         Inc(k);
      end;

      ll := ll + Dist(x1,y1,x2,y2);
      x1 := x2 ;
      y1 := y2 ;
   end;
   tLnsN := k ;
   SetLength(tLns, tLnsN);
end;



// _}[J̊Os𕡐ɕϊ
//    tLns, tLnsN 𗬗pAAł͂ȂƂĈ܂
procedure TransTenMarkerToLines(mc:integer;px,py,an,sc:double;t:TMatrix);
var
   i : integer ;
   sz,w1,w2,w3,w4,w5,w6 : double;
begin
   try
      tLnsN := 0 ;
      case(mc)of
      1,6: tLnsN :=  3*2 ;
      2,3: tLnsN := 24*2 ;
      4,7: tLnsN :=  2*2 ;
      5  : tLnsN :=  4*2 ;
      end;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   sz := 2.5/2.0*sc ;
   case(mc)of
   1: begin
         // asterisk
         w5 := px ;
         w6 := py+sz ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px ;
         w6 := py-sz ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;
         tLns[1].x := w3 ; tLns[1].y := w4 ;

         w5 := px+sz*Cos(30.0/180.0*Pi) ;
         w6 := py+sz*Sin(30.0/180.0*Pi) ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px+sz*Cos(210.0/180.0*Pi) ;
         w6 := py+sz*Sin(210.0/180.0*Pi) ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[2].x := w1 ; tLns[2].y := w2 ;
         tLns[3].x := w3 ; tLns[3].y := w4 ;

         w5 := px+sz*Cos(150.0/180.0*Pi) ;
         w6 := py+sz*Sin(150.0/180.0*Pi) ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px+sz*Cos(330.0/180.0*Pi) ;
         w6 := py+sz*Sin(330.0/180.0*Pi) ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[4].x := w1 ; tLns[4].y := w2 ;
         tLns[5].x := w3 ; tLns[5].y := w4 ;
      end;
   2: begin
         // circle
         w5 := 2.0*Pi/24 ;
         for i:=0 to 23 do begin
            w1 := px + sz*Cos(i    *w5);
            w2 := py + sz*Sin(i    *w5);
            w3 := px + sz*Cos((i+1)*w5);
            w4 := py + sz*Sin((i+1)*w5);
            MtxXY(w1,w2, w1,w2, t);
            MtxXY(w3,w4, w3,w4, t);
            tLns[i*2  ].x := w1 ; tLns[i*2  ].y := w2 ;
            tLns[i*2+1].x := w3 ; tLns[i*2+1].y := w4 ;
         end;
      end;
   3: begin
         // dot
         w5 := 2.0*Pi/24 ;
         for i:=0 to 23 do begin
            w1 := px + 0.1*sz*Cos(i    *w5);
            w2 := py + 0.1*sz*Sin(i    *w5);
            w3 := px + 0.1*sz*Cos((i+1)*w5);
            w4 := py + 0.1*sz*Sin((i+1)*w5);
            MtxXY(w1,w2, w1,w2, t);
            MtxXY(w3,w4, w3,w4, t);
            tLns[i*2  ].x := w1 ; tLns[i*2  ].y := w2 ;
            tLns[i*2+1].x := w3 ; tLns[i*2+1].y := w4 ;
         end;
      end;
   4: begin
         // plus
         w5 := px-sz ;
         w6 := py ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px+sz ;
         w6 := py ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;
         tLns[1].x := w3 ; tLns[1].y := w4 ;

         w5 := px ;
         w6 := py+sz ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px ;
         w6 := py-sz ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[2].x := w1 ; tLns[2].y := w2 ;
         tLns[3].x := w3 ; tLns[3].y := w4 ;
      end;
   5: begin
         // square
         w5 := px-sz ;
         w6 := py-sz ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px+sz ;
         w6 := py-sz ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;
         tLns[1].x := w3 ; tLns[1].y := w4 ;

         w5 := px+sz ;
         w6 := py-sz ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px+sz ;
         w6 := py+sz ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[2].x := w1 ; tLns[2].y := w2 ;
         tLns[3].x := w3 ; tLns[3].y := w4 ;

         w5 := px+sz ;
         w6 := py+sz ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px-sz ;
         w6 := py+sz ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[4].x := w1 ; tLns[4].y := w2 ;
         tLns[5].x := w3 ; tLns[5].y := w4 ;

         w5 := px-sz ;
         w6 := py+sz ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px-sz ;
         w6 := py-sz ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[6].x := w1 ; tLns[6].y := w2 ;
         tLns[7].x := w3 ; tLns[7].y := w4 ;
      end;
   6: begin
         // triangle
         sz := 2.5*2.0/3.0*sc ;
         w5 := px+sz*Cos(90.0/180.0*Pi) ;
         w6 := py+sz*Sin(90.0/180.0*Pi) ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px+sz*Cos(210.0/180.0*Pi) ;
         w6 := py+sz*Sin(210.0/180.0*Pi) ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;
         tLns[1].x := w3 ; tLns[1].y := w4 ;

         w5 := px+sz*Cos(210.0/180.0*Pi) ;
         w6 := py+sz*Sin(210.0/180.0*Pi) ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px+sz*Cos(330.0/180.0*Pi) ;
         w6 := py+sz*Sin(330.0/180.0*Pi) ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[2].x := w1 ; tLns[2].y := w2 ;
         tLns[3].x := w3 ; tLns[3].y := w4 ;

         w5 := px+sz*Cos(330.0/180.0*Pi) ;
         w6 := py+sz*Sin(330.0/180.0*Pi) ;
         Rev(w1,w2, w5,w6, px,py, an);
         w5 := px+sz*Cos(90.0/180.0*Pi) ;
         w6 := py+sz*Sin(90.0/180.0*Pi) ;
         Rev(w3,w4, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[4].x := w1 ; tLns[4].y := w2 ;
         tLns[5].x := w3 ; tLns[5].y := w4 ;
      end;
   7: begin
         // X
         w5 := px-sz ;
         w6 := py ;
         Rev(w1,w2, w5,w6, px,py, an+0.25*Pi);
         w5 := px+sz ;
         w6 := py ;
         Rev(w3,w4, w5,w6, px,py, an+0.25*Pi);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;
         tLns[1].x := w3 ; tLns[1].y := w4 ;

         w5 := px ;
         w6 := py+sz ;
         Rev(w1,w2, w5,w6, px,py, an+0.25*Pi);
         w5 := px ;
         w6 := py-sz ;
         Rev(w3,w4, w5,w6, px,py, an+0.25*Pi);
         MtxXY(w1,w2, w1,w2, t);
         MtxXY(w3,w4, w3,w4, t);
         tLns[2].x := w1 ; tLns[2].y := w2 ;
         tLns[3].x := w3 ; tLns[3].y := w4 ;
      end;
   end;
end;

// ̊Os𕡐ɕϊ
//    tLns, tLnsN 𗬗pAAł͂ȂƂĈ܂
procedure TransArrowToLines(ac:integer;px,py,an,sc:double;t:TMatrix);
var
   sx : double ;
   w1,w2,w3,w5,w6 : double ;
   i : integer ;
begin
   try
      tLnsN := 0 ;
      case(ac)of
      1,6,11   : tLnsN :=  4 ;
      2,5      : tLnsN :=  5 ;
      3,4,7,8  : tLnsN := 25 ;
      9        : tLnsN :=  3 ;
      10       : tLnsN :=  2 ;
      end;
      SetLength(tLns, tLnsN);
   except
      TransLnsClear ;
      exit ;
   end;

   case(ac)of
   1,6,11:
      begin    // blanked arrow,filled arrow,unfilled arrow
         w1 := px ;
         w2 := py ;
         MtxXY(w1,w2, w1,w2, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;
         tLns[3].x := w1 ; tLns[3].y := w2 ;

         w5 := px + 7.5 *sc ;
         w6 := py + 1.25*sc ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[1].x := w1 ; tLns[1].y := w2 ;

         w5 := px + 7.5 *sc ;
         w6 := py - 1.25*sc ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[2].x := w1 ; tLns[2].y := w2 ;
      end;
   2,5:
      begin    // blanked box,filled box
         w5 := px - 1.25*sc ;
         w6 := py - 1.25*sc ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;
         tLns[4].x := w1 ; tLns[4].y := w2 ;

         w5 := px + 1.25*sc ;
         w6 := py - 1.25*sc ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[1].x := w1 ; tLns[1].y := w2 ;

         w5 := px + 1.25*sc ;
         w6 := py + 1.25*sc ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[2].x := w1 ; tLns[2].y := w2 ;

         w5 := px - 1.25*sc ;
         w6 := py + 1.25*sc ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[3].x := w1 ; tLns[3].y := w2 ;
      end;
   3,4,7:
      begin    // blanked dot,dimension origin,filled dot
         w3 := 2.0*Pi/24 ;
         for i:=0 to 24 do begin
            w1 := px + 1.25*sc*Cos(i*w3);
            w2 := py + 1.25*sc*Sin(i*w3);
            MtxXY(w1,w2, w1,w2, t);
            tLns[i].x := w1 ; tLns[i].y := w2 ;
         end;
      end;
   8: begin    // integral symbol
         sx := 2.5*Sqrt(2.0);
         w3 := Pi/2.0/12 ;
         for i:=0 to 12 do begin
            w5 := px-sx/2.0 + sx/2.0*Cos(1.5*Pi+i*w3);
            w6 := py        + sx/2.0*Sin(1.5*Pi+i*w3);
            Rev(w1,w2, w5,w6, px,py, an);
            MtxXY(w1,w2, w1,w2, t);
            tLns[i].x := w1 ; tLns[i].y := w2 ;
         end;
         for i:=1 to 12 do begin
            w5 := px+sx/2.0 + sx/2.0*Cos(Pi-i*w3);
            w6 := py        + sx/2.0*Sin(Pi-i*w3);
            Rev(w1,w2, w5,w6, px,py, an);
            MtxXY(w1,w2, w1,w2, t);
            tLns[i+12].x := w1 ; tLns[i+12].y := w2 ;
         end;
      end;
   9: begin    // open arrow
         w5 := px + 7.5 *sc ;
         w6 := py + 1.25*sc ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;

         w1 := px ;
         w2 := py ;
         MtxXY(w1,w2, w1,w2, t);
         tLns[1].x := w1 ; tLns[1].y := w2 ;

         w5 := px + 7.5 *sc ;
         w6 := py - 1.25*sc ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[2].x := w1 ; tLns[2].y := w2 ;
      end;
   10:begin    // slash
         sx := 2.5*Sqrt(2.0);
         w5 := px - sx/2.0 ;
         w6 := py - sx/2.0 ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[0].x := w1 ; tLns[0].y := w2 ;

         w5 := px + sx/2.0 ;
         w6 := py + sx/2.0 ;
         Rev(w1,w2, w5,w6, px,py, an);
         MtxXY(w1,w2, w1,w2, t);
         tLns[1].x := w1 ; tLns[1].y := w2 ;
      end;
   end;
end;

end.
